// This file was generated by ecm_setup_version(): DO NOT EDIT!

#ifndef KWINDOWSYSTEM_VERSION_H
#define KWINDOWSYSTEM_VERSION_H

#define KWINDOWSYSTEM_VERSION_STRING "6.24.0"
#define KWINDOWSYSTEM_VERSION_MAJOR 6
#define KWINDOWSYSTEM_VERSION_MINOR 24
#define KWINDOWSYSTEM_VERSION_PATCH 0
#define KWINDOWSYSTEM_VERSION ((6<<16)|(24<<8)|(0))

#endif
